;Nixie Clock (with new DCF77 receiver)
;Controller: ATTiny2313 @ 4.194304 MHz
;
;Assemble with avra (http://avra.sourceforge.net/) only, this file contains
;directives specific for this assembler!
;
;(C) Arne Rossius, http://www.elektronik.de.vu
;
;PORTS:
;  PB6 = Shift Register Data (MSB first, seconds first, packed BCD)
;  PB7 = Shift Register Clock
;  PD2 = DCF77 receiver
;  PD3 = "Show Date" key/switch
;
;VERSION HISTORY:
;  0.1  2005-09-23  Arne Rossius
;  0.2  2007-05-20  Arne Rossius
;    * auto-increment date
;  0.3  2008-01-12  Arne Rossius
;    * date format selectable (dd-mm-yy or yy-mm-dd)
;  0.4  2008-03-15  Arne Rossius
;    * auto-increment weekday (for future use)
;    * disable DCF77 reception after 4 minutes without success
;  0.5  2008-04-22  Arne Rossius
;    * BUGFIX: auto-increment date: load ZH:ZL instead of YH:ZL (narf!)

;microcontroller definition file
.include "../include/2313def.inc"

;Nixie clock options
.equ	DATE_TIMEOUT = 5 ;seconds
.equ	DD_MM_YY = 0 ;German date format instead of international

;DCF77 receiver options
.equ	DCFPort = PORTD
.equ	DCFPin = 2
.equ	DCFInvert = 1 ;inverted output (negative pulses), most receivers do this!

;DCF77 reception options
.equ	DCFReceiveEveryHour = 1 ;receive only every hour (instead of all the time)
.equ	DCFCheckStartBit = 1 ;check whether the start bit is 1
.equ	DCFNarrowBitLength = 1 ;don't allow bit lengths that are far from 100ms or 200ms
.equ	DCFStopOnError = 1 ;stop DCF77 reception if an error was encountered until the next telegram begins
.equ	DCFStoreErrors = 1 ;store the last error in the flags register (setting this to 0 implies DCFStopOnError = 0)

;registers for DCF77 telegram
.def	dcf1 = R3
.def	dcf2 = R4
.def	dcf3 = R5
.def	dcf4 = R6
.def	dcf5 = R7

;time/date registers
.def	subsec = R8
.def	second = R9
.def	minute = R10
.def	hour = R11
.def	day = R12
.def	month = R13
.def	year = R14
.def	wkday = R15 ;(weekday, 1=MO 7=SU)

;general registers
.def	temp = R16
.def	temp2 = R17

;DCF77-specific registers
.def	dcfbits = R18 ;received bits
.def	dcftime = R19 ;1/64-seconds since last edge
.def	flags1 = R20
.def	flags2 = R21

;nixie clock registers
.def	oldout1 = R22
.def	oldout2 = R23
.def	oldout3 = R24
.def	newout1 = R25
.def	newout2 = R26
.def	newout3 = R27
.def	datecnt = R28

;bit definitions for the flags1 register
.equ	ErrBitCount = 0 ;last error: wrong number of bits
.equ	ErrParity = 1 ;last error: parity error
.equ	ErrBitLength = 2 ;last error: wrong bit length
.equ	ErrStartBit = 3 ;last error: start bit was not 1
.equ	TimeOK = 4 ;time is OK
.equ	DateOK = 5 ;date is OK (reset at midnight)
.equ	StopDCF = 6 ;stop DCF77 reception until a new telegram starts
.equ	DCFdisabled = 7 ;DCF77 receiver disabled (reset when a new hour begins)
.equ	ErrMask = 1<<ErrBitLength | 1<<ErrParity | 1<<ErrBitCount | 1<<ErrStartBit ;used to reset errors

;bit definitions for the flags2 register
.equ	LastState = DCFPin ;last state of DCF77 input, must be the same bit as on the port!

.cseg

;*******************************************************************************
;*
;*   MAIN PROGRAM
;*
;*******************************************************************************

;---------------------------------------
; Interrupt vectors
;---------------------------------------

;Reset
.org $0000
	rjmp	start
	
;Timer 0 overflow
.org OVF0addr
	rjmp	ovf0int

;---------------------------------------
; Length of every month in days
;---------------------------------------

months:
	;length of months + 1, packed BCD
	;    Jan   Feb   Mar   Apr   May   Jun   Jul   Aug   Sep   Oct   Nov   Dec
	.db 0x32, 0x29, 0x32, 0x31, 0x32, 0x31, 0x32, 0x32, 0x31, 0x32, 0x31, 0x32 ;normal
	.db 0x32, 0x30, 0x32, 0x31, 0x32, 0x31, 0x32, 0x32, 0x31, 0x32, 0x31, 0x32 ;leap year

;---------------------------------------
; Initialization
;---------------------------------------
	
start:
	;stack pointer
	ldi	temp, LOW(RAMEND)
	out	SPL, temp
	
	;disable analog comparator
	ldi	temp, 0x80
	out	ACSR, temp
	
	;I/O ports
	ldi	temp, 0xC0
	out	DDRB, temp
	ldi	temp, 0x3F
	out	PORTB, temp
	ldi	temp, 0x00
	out	DDRD, temp
	ldi	temp, 0x7F
	out	PORTD, temp
	
	;timer
	ldi	temp, 1<<TOIE0 ;overflow int on
	out	TIMSK, temp
	ldi	temp, 0x04 ;Clk/256 = 64Hz int
	out	TCCR0, temp
	
	;registers
	ldi	dcfbits, 0
	ldi	dcftime, 0
	ldi	flags1, 1<<StopDCF
	ldi	flags2, 0
	
	ldi	datecnt, 0
	ldi	oldout1, 0x00
	ldi	oldout2, 0x00
	ldi	oldout3, 0x00
	ldi	newout1, 0x88
	ldi	newout2, 0x88
	ldi	newout3, 0x88
	
	;enable interrupts
	sei

;---------------------------------------
; Main loop
;   find out what must be displayed and
;   shift it out if the current display
;   is different
;---------------------------------------

loop:
	;check key
	sbic	PIND, 3
	rjmp	key_off
	;key is pressed
	cpi	datecnt, 0
	brne	check_display
	ldi	datecnt, DATE_TIMEOUT*2
	rjmp	check_display
key_off:
	;key is not pressed
	cpi	datecnt, DATE_TIMEOUT+1
	brsh	check_display
	clr	datecnt
	
check_display:
	;display date?
	sbrs	flags1, DateOK
	rjmp	date_err
	cpi	datecnt, DATE_TIMEOUT+1
	brsh	date
date_err:
	;display time?
	sbrc	flags1, TimeOK
	rjmp	time
	;display DCF77 reception info (88:x8:yy, x=error, yy=received bits)
	ldi	newout1, 0x88
	ldi	newout2, 0x08
	sbrc	flags1, ErrBitCount
	ori	newout2, 0x10
	sbrc	flags1, ErrBitLength
	ori	newout2, 0x20
	sbrc	flags1, ErrParity
	ori	newout2, 0x30
	sbrc	flags1, ErrStartBit
	ori	newout2, 0x40
	mov	temp, dcfbits
	rcall	bin2bcd
	mov	newout3, temp
	
display:
	;check if the display has changed:
	cp	oldout1, newout1
	cpc	oldout2, newout2
	cpc	oldout3, newout3
	breq	loop
	;display the new data
	mov	temp, newout3
	rcall	shiftbyte
	mov	temp, newout2
	rcall	shiftbyte
	mov	temp, newout1
	rcall	shiftbyte
	;save the new display
	mov	oldout1, newout1
	mov	oldout2, newout2
	mov	oldout3, newout3
	rjmp	loop
	
date:
	;display date
.if (DD_MM_YY == 1)
	mov	newout1, day
.else
	mov	newout1, year
.endif
	mov	newout2, month
.if (DD_MM_YY == 1)
	mov	newout3, year
.else
	mov	newout3, day
.endif
	rjmp	display
	
time:
	;display time
	mov	newout1, hour
	mov	newout2, minute
	mov	newout3, second
	rjmp	display
	
;*******************************************************************************
;*
;*   Nixie clock subroutines
;*
;*******************************************************************************

;---------------------------------------
; Shift out one byte (MSB first)
;   in: temp
;   modifies: temp, temp2
;---------------------------------------

shiftbyte:
	ldi	temp2, 8
shiftbyte_loop:
	cbi	PORTB, 6
	lsl	temp
	brcc	PC+2
	sbi	PORTB, 6
	sbi	PORTB, 7
	nop
	cbi	PORTB, 7
	dec	temp2
	brne	shiftbyte_loop
	ret

;---------------------------------------
; convert binary to packed BCD
;   in: temp
;   out: temp
;   modifies: temp2
;---------------------------------------

bin2bcd:
	ldi	temp2, 0
bin2bcd_loop:
	subi	temp2, -0x10
	subi	temp, 10
	brcc	bin2bcd_loop
	subi	temp2, 0x10
	subi	temp, -10
	add	temp, temp2
	ret
	
;*******************************************************************************
;*
;*   DCF77 subroutines
;*
;*******************************************************************************

;---------------------------------------
; Timer 0 overflow int (64Hz)
;   modifies: dcftime, dcfbits, flags1, flags2, dcf1, dcf2, dcf3, dcf4, dcf5,
;             subsec, second, minute, hour, day, wkday, month, year
;---------------------------------------

ovf0int:
	;save registers
	push	temp
	push	temp2
	in	temp, SREG
	push	temp
	
	.if ( DCFReceiveEveryHour )
		;don't evaluate DCF77 signal when reception is disabled
		sbrc	flags1, DCFdisabled
		rjmp	dcf_end
		.message "Note: DCF77 reception: every hour"
	.else
		.message "Note: DCF77 reception: all the time"
	.endif
	
	inc	dcftime
	
	;compare DCF77 input to last saved state
	in	temp, DCFPort-2 ;PINx
	eor	temp, flags2
	andi	temp, 1<<LastState
	.if ( DCFInvert )
		.message "Note: DCF77 reception: assembling for inverted DCF77 signal"
		brne	dcf_end ;no change
	.else
		.message "Note: DCF77 reception: assembling for non-inverted DCF77 signal"
		breq	dcf_end ;no change
	.endif
	
	;take appropriate action for pos/neg edges
	sbrs	flags2, LastState
	rjmp	dcf_posedge
	
		; < negative edge ( 1 -> 0 ) >
	
	andi	flags2, ~(1<<LastState)
	
	.if ( DCFStopOnError )
		.message "Note: DCF77 reception: will be stopped on error"
		.if ( DCFStoreErrors )
			;exit if an error was detected before
			sbrc	flags1, StopDCF
			rjmp	dcf_end
		.endif
	.else
		.message "Note: DCF77 reception: will CONTINUE on error"
	.endif
	 
	;increment bit number, exit if more than 59 bits were already received
	inc	dcfbits
	cpi	dcfbits, 60
	brsh	PC+2
	rjmp	dcf_evalbit
	.if ( DCFStoreErrors )
		.message "Note: DCF77 reception: errors will be stored"
		andi	flags1, ~ErrMask
		ori	flags1, 1<<ErrBitCount | 1<<StopDCF
	.else
		.message "Note: DCF77 reception: errors will NOT be stored"
	.endif
	rjmp	dcf_end
	
		; < positive edge ( 0 -> 1 ) >
dcf_posedge:
	ori	flags2, 1<<LastState
	
	;check if time was >1.2s (start of new telegram)
	cpi	dcftime, 77
	ldi	dcftime, 0
	brlo	dcf_end
	
	;end of telegram / new telegram
	.if ( DCFStoreErrors )
		andi	flags1, ~ErrMask & ~(1<<StopDCF)
	.endif
	cpi	dcfbits, 59
	ldi	dcfbits, 0
	brne	PC+2
	rjmp	dcf_eval
	.if ( DCFStoreErrors )
		ori	flags1, 1<<ErrBitCount
	.endif
dcf_end:

;---------------------------------------
; increment time (fallback quartz clock)
;---------------------------------------

	;increment 1/64-seconds
	inc	subsec
	ldi	temp, 64
	cp	subsec, temp
	brlo	inctime_end
	
	;increment seconds
	clr	subsec
	mov	temp, second
	rcall	bcdinc
	mov	second, temp
	cpi	datecnt, 0 ;nixieclock: decrement date display timeout
	breq	PC+2
	dec	datecnt
	cpi	temp, 0x60
	brlo	inctime_end
	
	;increment minutes
	clr	second
	mov	temp, minute
	rcall	bcdinc
	mov	minute, temp
	cpi	temp, 0x04 ;disable DCF77 reception if nothing received for 4 minutes
	brne	inctime_nodisable
	sbrc	flags1, TimeOK
	ori	flags1, 1<<DCFdisabled
inctime_nodisable:
	cpi	temp, 0x60
	brlo	inctime_end

	;increment hours
	clr	minute
	.if ( DCFReceiveEveryHour )
		andi	flags1, LOW(~(1<<DCFdisabled)) ;re-enable DCF77 reception
		;               ^^^ workaround for avra warning on ~0x80
	.endif
	mov	temp, hour
	rcall	bcdinc
	mov	hour, temp
	cpi	temp, 0x24
	brlo	inctime_end
	
	;increment weekday and days
	clr	hour
	mov	temp, wkday
	inc	temp
	cpi	temp, 8
	brlo	PC+2
	ldi	temp, 1
	mov	wkday, temp
	ldi	ZL, LOW(months*2-1) ;get length of current month
	ldi	ZH, HIGH(months*2-1)
	sbrc	year, 0
	rjmp	noleap
	sbrc	year, 1
	rjmp	noleap
	adiw	ZH:ZL, 12
noleap:
	mov	temp, month
	rcall	bcd2bin
	add	ZL, temp
	ldi	temp, 0
	adc	ZH, temp
	lpm
	mov	temp, day ;increment day
	rcall	bcdinc
	mov	day, temp
	cp	temp, R0 ;R0 is result from LPM
	brlo	inctime_end
	
	;increment month
	clr	day
	inc	day
	mov	temp, month
	rcall	bcdinc
	mov	month, temp
	cpi	temp, 0x13
	brlo	inctime_end
	
	;increment year
	clr	month
	inc	month
	mov	temp, year
	rcall	bcdinc
	mov	year, temp
inctime_end:

	pop	temp
	out	SREG, temp
	pop	temp2
	pop	temp
	reti

;---------------------------------------
; DCF77 reception (continued)
;---------------------------------------

		; < negative edge - cont'd >
	
	;check bit length
dcf_evalbit:
	.if ( DCFNarrowBitLength )
		.message "Note: DCF77 reception: narrow bit length check enabled"
		cpi	dcftime, 5 ;bit is 0: 5 <= dcftime <= 7
		.if ( DCFStoreErrors )
			brlo	dcf_errbit
		.else
			brlo	dcf_end
		.endif
		cpi	dcftime, 8
		brlo	dcf_clearbit
		cpi	dcftime, 11 ;bit is 1: 11 <= dcftime <= 13
		.if ( DCFStoreErrors )
			brlo	dcf_errbit
		.else
			brlo	dcf_end
		.endif
		cpi	dcftime, 14
		brlo	dcf_setbit
	
		.if ( DCFStoreErrors )
			;received bit is erroneous
		dcf_errbit:
			andi	flags1, ~ErrMask
			ori	flags1, 1<<ErrBitLength | 1<<StopDCF
			rjmp	dcf_end
		.endif
	.else
		.message "Note: DCF77 reception: bit length check DISABLED"
		cpi	dcftime, 10
		brlo	dcf_clearbit
	.endif
	
	;save bit
dcf_setbit:
	sec ;set bit
	rjmp	dcf_savebit
dcf_clearbit:
	clc ;clear bit
dcf_savebit:
	ror	dcf5
	ror	dcf4
	ror	dcf3
	ror	dcf2
	ror	dcf1
	rjmp	dcf_end
	
;---------------------------------------
; evaluate DCF77 telegram
;---------------------------------------
	
dcf_eval:
	.if ( DCFCheckStartBit )
		.message "Note: DCF77 evaluation: start bit will be checked"
		;check start bit
		sbrc	dcf1, 1
		rjmp	dcf_startbit_ok
		andi	flags1, ~ErrMask & ~(1<<StopDCF)
		ori	flags1, 1<<ErrStartBit | 1<<StopDCF
		rjmp	dcf_end
	dcf_startbit_ok:
	.else
		.message "Note: DCF77 evaluation: start bit will NOT be checked"
	.endif
	
	;check date parity
	ldi	temp2, 0
	mov	temp, dcf3
	andi	temp, 0xFE
	rcall	parity
	mov	temp, dcf4
	rcall	parity
	mov	temp, dcf5
	andi	temp, 0x7F
	rcall	parity
	mov	temp, dcf5
	lsl	temp ;Bit7 -> C
	rol	temp ;C -> Bit0
	andi	temp, 0x01
	cp	temp, temp2
	brne	dcf_parityerror
	
	;check hour parity
	ldi	temp2, 0
	mov	temp, dcf2
	andi	temp, 0xFC
	rcall	parity
	mov	temp, dcf3
	andi	temp, 0x01
	cp	temp, temp2
	brne	dcf_parityerror
	
	;check minute parity
	ldi	temp2, 0
	mov	temp, dcf1
	andi	temp, 0xFC
	rcall	parity
	mov	temp, dcf2
	andi	temp, 0x01
	rcall	parity
	mov	temp, dcf2
	lsr	temp
	andi	temp, 0x01
	cp	temp, temp2
	brne	dcf_parityerror

	;save minutes
	lsr	dcf2
	ror	dcf1
	lsr	dcf1
	mov	minute, dcf1
	
	;save hours
	lsr	dcf2
	mov	hour, dcf2
	
	;save day
	lsl	dcf3
	rol	dcf4 ;align month[5] weekday[3]
	rol	dcf5 ;align year
	lsr	dcf3 ;align day
	lsr	dcf3
	mov	day, dcf3
	
	;save weekday
	mov	wkday, dcf4
	ldi	temp, 0x07
	and	wkday, temp
	
	;save month
	lsr	dcf4
	lsr	dcf4
	lsr	dcf4
	mov	month, dcf4
	
	;save year
	mov	year, dcf5
	
	;set time & date ok, disable DCF77 reception
	clr	subsec
	clr	second
	ori	flags1, 1<<TimeOK | 1<<DateOK | 1<<DCFdisabled
	rjmp	dcf_end
	
dcf_parityerror:
	.if ( DCFStoreErrors )
		;set parity error
		ori	flags1, 1<<ErrParity
	.endif
	rjmp	dcf_end

;*******************************************************************************
;*
;*   GENERAL ARITHMETIC FUNCTIONS
;*
;*******************************************************************************

;---------------------------------------
; calculate parity and add up
;   in: temp
;   out: temp2 (add to existing value)
;---------------------------------------

parity:
	lsr	temp
	brcc	PC+2
	inc	temp2
	cpi	temp, 0
	brne	parity
	andi	temp2, 0x01
	ret

;---------------------------------------
; increment packed bcd number
;   in: temp
;   out: temp
;---------------------------------------

bcdinc:
	subi	temp, -7 ;increment and add 6
	brhc	PC+2 ;half-carry is inverted when using "subi rh, -k" instead of add!
	subi	temp, 6
	ret
	
;---------------------------------------
; convert packed BCD to binary
;   in: temp
;   out: temp
;   modifies: temp2
;---------------------------------------

bcd2bin:
	mov	temp2, temp
	andi	temp, 0x0F
	swap	temp2
	andi	temp2, 0x0F
	lsl	temp2
	add	temp, temp2 ;temp += temp2*2
	lsl	temp2
	lsl	temp2
	add	temp, temp2 ;temp += temp2*8
	ret
